"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GitbookDownloader = void 0;
const jsdom_1 = require("jsdom");
const turndown_1 = __importDefault(require("turndown"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const pino_1 = __importDefault(require("pino"));
const logger = (0, pino_1.default)({
    transport: {
        target: 'pino-pretty',
        options: {
            colorize: true,
            translateTime: 'HH:MM:ss',
            ignore: 'pid,hostname'
        }
    }
});
/**
 * GitbookDownloader class for converting Gitbook pages to markdown
 */
class GitbookDownloader {
    /**
     * Creates a new GitbookDownloader instance
     * @param baseUrl - The base URL of the Gitbook documentation
     * @param options - Configuration options
     */
    constructor(baseUrl, options = {}) {
        this.baseUrl = baseUrl;
        this.options = {
            delay: 1000,
            maxRetries: 3,
            retryDelay: 5000,
            headers: {},
            outputPath: 'output.md',
            maxConcurrency: 10,
            ...options
        };
        this.status = {
            totalPages: 0,
            currentPage: 0,
            currentUrl: '',
            status: 'idle',
            pagesScraped: [],
        };
        this.visitedUrls = new Set();
        this.pages = new Map();
        this.contentHash = new Map();
        this.turndownService = new turndown_1.default({
            headingStyle: 'atx',
            codeBlockStyle: 'fenced',
            hr: '---',
            bulletListMarker: '*'
        });
        // Configure TurndownService
        this.turndownService.addRule('codeBlocks', {
            filter: ['pre', 'code'],
            replacement: function (content, node) {
                const element = node;
                const language = element.getAttribute('class')?.replace('language-', '') || '';
                return `\n\`\`\`${language}\n${content}\n\`\`\`\n`;
            }
        });
        // Add rule for tables
        this.turndownService.addRule('tables', {
            filter: ['table'],
            replacement: function (content) {
                return '\n\n' + content + '\n\n';
            }
        });
    }
    /**
     * Downloads and converts Gitbook content to markdown
     */
    async download() {
        try {
            this.status.startTime = Date.now();
            this.status.status = 'downloading';
            this.status.currentUrl = this.baseUrl;
            // Try to fetch sitemap first
            logger.info('Starting download process');
            logger.debug('Attempting to fetch sitemap');
            const sitemapUrls = await this._fetchSitemap();
            let urls = [];
            if (sitemapUrls.length > 0) {
                logger.info({ count: sitemapUrls.length }, 'Found URLs in sitemap');
                urls = sitemapUrls;
            }
            else {
                // Fallback to manual navigation extraction
                logger.info('No sitemap found, falling back to navigation extraction');
                const initialContent = await this._fetchPage(this.baseUrl);
                if (!initialContent) {
                    throw new Error('Failed to fetch main page');
                }
                const dom = new jsdom_1.JSDOM(initialContent);
                urls = this._extractNavLinks(dom.window.document);
                logger.info({ count: urls.length }, 'Found URLs through navigation');
                // Add the main page if not already included
                if (!urls.includes(this.baseUrl)) {
                    urls.unshift(this.baseUrl);
                    logger.debug('Added main page URL to processing list');
                }
            }
            this.status.totalPages = urls.length;
            logger.info({ totalPages: this.status.totalPages }, 'Starting page processing');
            // Process pages in parallel with rate limiting
            const chunks = this._chunkArray(urls, this.options.maxConcurrency);
            logger.debug({ chunks: chunks.length, chunkSize: this.options.maxConcurrency }, 'Split URLs into chunks');
            for (const chunk of chunks) {
                logger.debug({ urls: chunk.length }, 'Processing chunk of URLs');
                await Promise.all(chunk.map(url => this._processUrl(url)));
                logger.trace({ delay: this.options.delay }, 'Waiting before next chunk');
                await this._sleep(this.options.delay);
            }
            logger.info({ processed: this.pages.size }, 'Completed page processing');
            const markdown = await this._generateMarkdown();
            this.status.status = 'completed';
            logger.info('Download completed successfully');
            return markdown;
        }
        catch (error) {
            this.status.status = 'error';
            this.status.error = error.message;
            logger.error({ err: error }, 'Download failed');
            throw error;
        }
    }
    /**
     * Attempts to fetch and parse the sitemap
     */
    async _fetchSitemap() {
        const urls = [];
        const sitemapPaths = [
            '/sitemap.xml',
            '/docs/sitemap.xml',
            '/documentation/sitemap.xml'
        ];
        for (const path of sitemapPaths) {
            try {
                const sitemapUrl = new URL(path, this.baseUrl).toString();
                logger.debug({ url: sitemapUrl }, 'Trying sitemap');
                const response = await (0, node_fetch_1.default)(sitemapUrl);
                if (response.ok) {
                    const content = await response.text();
                    const dom = new jsdom_1.JSDOM(content, { contentType: 'text/xml' });
                    const locations = dom.window.document.querySelectorAll('loc');
                    locations.forEach(loc => {
                        const url = loc.textContent;
                        if (url && url.startsWith(this.baseUrl)) {
                            urls.push(url);
                        }
                    });
                    if (urls.length > 0) {
                        logger.info({ count: urls.length, sitemap: sitemapUrl }, 'Found URLs in sitemap');
                        break;
                    }
                }
            }
            catch (error) {
                logger.debug({ path }, 'No sitemap found');
            }
        }
        return urls;
    }
    /**
     * Processes a single URL
     */
    async _processUrl(url) {
        if (this.visitedUrls.has(url)) {
            logger.debug({ url }, 'Skipping already visited URL');
            return;
        }
        logger.debug({ url }, 'Processing URL');
        this.visitedUrls.add(url);
        const content = await this._fetchPage(url);
        if (content) {
            await this._processPageContent(content, url);
            this.status.currentPage++;
            logger.trace({ current: this.status.currentPage, total: this.status.totalPages }, 'Page processed');
        }
    }
    /**
     * Splits an array into chunks for parallel processing
     */
    _chunkArray(array, size) {
        const chunks = [];
        for (let i = 0; i < array.length; i += size) {
            chunks.push(array.slice(i, i + size));
        }
        return chunks;
    }
    /**
     * Gets the current download status
     */
    getStatus() {
        return { ...this.status };
    }
    /**
     * Fetches a page and handles retries
     */
    async _fetchPage(url) {
        let retries = 0;
        while (retries < this.options.maxRetries) {
            try {
                logger.debug('Fetching %s', url);
                const response = await (0, node_fetch_1.default)(url, {
                    headers: {
                        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko)',
                        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
                        'Accept-Language': 'en-US,en;q=0.5',
                        ...this.options.headers
                    }
                });
                if (response.status === 429) {
                    const reset = response.headers.get('x-ratelimit-reset');
                    if (reset) {
                        this.status.rateLimitReset = parseInt(reset) * 1000;
                        logger.warn('Rate limit hit, waiting until %s', new Date(this.status.rateLimitReset).toISOString());
                        await this._sleep(this.status.rateLimitReset - Date.now());
                    }
                    continue;
                }
                if (!response.ok) {
                    throw new Error(`HTTP error! status: ${response.status}`);
                }
                const content = await response.text();
                logger.trace('Fetched %d bytes from %s', content.length, url);
                await this._sleep(this.options.delay);
                return content;
            }
            catch (error) {
                logger.error('Error fetching %s: %o', url, error);
                retries++;
                if (retries === this.options.maxRetries) {
                    throw error;
                }
                logger.warn('Retry %d/%d after %d ms', retries, this.options.maxRetries, this.options.retryDelay);
                await this._sleep(this.options.retryDelay);
            }
        }
        return null;
    }
    /**
     * Processes page content and extracts navigation links
     */
    async _processPageContent(content, url) {
        const dom = new jsdom_1.JSDOM(content);
        const document = dom.window.document;
        // Extract title
        let title = '';
        const h1 = document.querySelector('h1');
        if (h1) {
            title = h1.textContent?.trim() || '';
        }
        if (!title) {
            const titleTag = document.querySelector('title');
            if (titleTag) {
                title = titleTag.textContent?.split('|')[0]?.trim() || '';
            }
        }
        if (!title) {
            title = new URL(url).pathname.split('/').pop()?.replace(/-/g, ' ') || 'Introduction';
        }
        logger.debug('Processing page: %s', title);
        // Get main content
        let mainContent = document.querySelector('main');
        if (!mainContent) {
            mainContent = document.querySelector('article');
        }
        if (!mainContent) {
            mainContent = document.querySelector('[class*="markdown"], [class*="content"], [class*="article"], [class*="documentation"]');
        }
        if (!mainContent) {
            // Look for the largest content block
            const contentBlocks = Array.from(document.querySelectorAll('div')).filter(div => {
                const text = div.textContent || '';
                return text.length > 500 && !div.querySelector('nav, header, footer');
            });
            mainContent = contentBlocks.sort((a, b) => (b.textContent?.length || 0) - (a.textContent?.length || 0))[0];
        }
        if (!mainContent) {
            mainContent = document.body;
        }
        logger.trace('Found main content element: %s', mainContent.tagName);
        // Clone the content to avoid modifying the original
        const contentClone = mainContent.cloneNode(true);
        // Remove navigation elements
        const removedNav = contentClone.querySelectorAll('nav, aside, header, footer, [class*="navigation"], [class*="sidebar"], [class*="menu"]');
        logger.trace('Removed %d navigation elements', removedNav.length);
        removedNav.forEach(el => el.remove());
        // Remove scripts and styles
        const removedScripts = contentClone.querySelectorAll('script, style, iframe, noscript');
        logger.trace('Removed %d script/style elements', removedScripts.length);
        removedScripts.forEach(el => el.remove());
        // Remove navigation links at bottom
        const removedLinks = Array.from(contentClone.querySelectorAll('a')).filter(link => {
            const text = link.textContent?.toLowerCase() || '';
            return text.includes('previous') || text.includes('next') || text.includes('up') || text.includes('down');
        });
        logger.trace('Removed %d navigation links', removedLinks.length);
        removedLinks.forEach(link => link.parentElement?.remove());
        // Clean up content
        this._cleanupContent(contentClone);
        // Store page data
        const pageIndex = this.pages.size;
        const content_hash = this._hashString(contentClone.innerHTML);
        // Only store if content is unique
        if (!this.contentHash.has(content_hash)) {
            logger.debug('Storing page %d: %s', pageIndex, title);
            this.pages.set(pageIndex, {
                index: pageIndex,
                title,
                content: contentClone.innerHTML,
                url
            });
            this.contentHash.set(content_hash, pageIndex);
            this.status.pagesScraped.push(title);
        }
        else {
            logger.debug('Skipping duplicate content for %s', title);
        }
    }
    /**
     * Clean up the content by removing empty elements and unnecessary attributes
     */
    _cleanupContent(element) {
        // Remove empty elements
        const emptyElements = element.querySelectorAll('p:empty, div:empty, span:empty');
        emptyElements.forEach(el => el.remove());
        // Remove unnecessary attributes
        const allElements = element.getElementsByTagName('*');
        for (const el of allElements) {
            const allowedAttrs = ['href', 'src', 'alt', 'title', 'class'];
            const attrs = el.attributes;
            for (let i = attrs.length - 1; i >= 0; i--) {
                const attr = attrs[i];
                if (!allowedAttrs.includes(attr.name)) {
                    el.removeAttribute(attr.name);
                }
            }
        }
        // Convert relative URLs to absolute
        const links = element.querySelectorAll('a[href]');
        links.forEach(link => {
            const href = link.getAttribute('href');
            if (href && !href.startsWith('http')) {
                link.setAttribute('href', new URL(href, this.baseUrl).toString());
            }
        });
        const images = element.querySelectorAll('img[src]');
        images.forEach(img => {
            const src = img.getAttribute('src');
            if (src && !src.startsWith('http')) {
                img.setAttribute('src', new URL(src, this.baseUrl).toString());
            }
        });
    }
    /**
     * Extracts navigation links from the page
     */
    _extractNavLinks(document) {
        const links = [];
        const processedUrls = new Set();
        // Helper function to process links
        const processLink = (href) => {
            if (href && !href.startsWith('#')) {
                try {
                    const absoluteUrl = new URL(href, this.baseUrl).toString();
                    // Only include links that are part of the same documentation
                    if (absoluteUrl.startsWith(this.baseUrl) && !processedUrls.has(absoluteUrl)) {
                        console.log(`Found navigation link: ${absoluteUrl}`);
                        links.push(absoluteUrl);
                        processedUrls.add(absoluteUrl);
                    }
                }
                catch (e) {
                    console.error(`Invalid URL: ${href}`);
                }
            }
        };
        // Find GitBook navigation elements
        const navElements = document.querySelectorAll('nav, aside, [class*="navigation"], [class*="sidebar"], [class*="menu"]');
        console.log(`Found ${navElements.length} navigation elements`);
        navElements.forEach(nav => {
            // Look for lists that typically contain navigation
            const navLists = nav.querySelectorAll('ol, ul');
            console.log(`Found ${navLists.length} navigation lists in element`);
            navLists.forEach(navList => {
                // Process list items in order
                const listItems = navList.querySelectorAll('li');
                console.log(`Found ${listItems.length} list items in list`);
                listItems.forEach(li => {
                    const link = li.querySelector('a[href]');
                    if (link) {
                        processLink(link.getAttribute('href'));
                    }
                });
            });
            // Also check direct links in nav elements
            nav.querySelectorAll('a[href]').forEach(link => {
                processLink(link.getAttribute('href'));
            });
        });
        // Check for GitBook-specific navigation elements
        document.querySelectorAll('[class*="page-list"] a[href], [class*="table-of-contents"] a[href]').forEach(link => {
            processLink(link.getAttribute('href'));
        });
        // Also check for next/prev navigation links
        document.querySelectorAll('a[aria-label], a[rel="prev"], a[rel="next"]').forEach(link => {
            const ariaLabel = link.getAttribute('aria-label')?.toLowerCase() || '';
            const rel = link.getAttribute('rel');
            if (ariaLabel.includes('next') || ariaLabel.includes('previous') || rel === 'prev' || rel === 'next') {
                processLink(link.getAttribute('href'));
            }
        });
        return links;
    }
    /**
     * Generates the final markdown content
     */
    async _generateMarkdown() {
        const sortedPages = Array.from(this.pages.values())
            .sort((a, b) => a.index - b.index);
        console.log(`Generating markdown for ${sortedPages.length} pages`);
        let markdown = '';
        for (const page of sortedPages) {
            console.log(`Converting page: ${page.title}`);
            markdown += `# ${page.title}\n\n`;
            markdown += this.turndownService.turndown(page.content);
            markdown += '\n\n---\n\n';
        }
        return markdown.trim();
    }
    /**
     * Utility method for sleeping
     */
    _sleep(ms) {
        return new Promise(resolve => setTimeout(resolve, ms));
    }
    /**
     * Utility method for hashing strings
     */
    _hashString(str) {
        let hash = 0;
        for (let i = 0; i < str.length; i++) {
            const char = str.charCodeAt(i);
            hash = ((hash << 5) - hash) + char;
            hash = hash & hash;
        }
        return hash.toString();
    }
}
exports.GitbookDownloader = GitbookDownloader;
